from _typeshed import Incomplete
from collections.abc import Generator
from typing import Tuple, Optional

import numpy as np

ThreeTuple = Tuple[float, float, float]
FourTuple = Tuple[float, float, float, float]


class Quaternion:
    q: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...
    def __hash__(self) -> int: ...
    @classmethod
    def random(cls) -> 'Quaternion': ...
    def __format__(self, formatstr: str) -> str: ...
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    def __complex__(self) -> complex: ...
    def __bool__(self) -> bool: ...
    def __nonzero__(self) -> bool: ...
    def __invert__(self) -> bool: ...
    def __eq__(self, other): ...
    def __neg__(self) -> 'Quaternion': ...
    def __abs__(self) -> float: ...
    def __add__(self, other): ...
    def __iadd__(self, other): ...
    def __radd__(self, other): ...
    def __sub__(self, other): ...
    def __isub__(self, other): ...
    def __rsub__(self, other): ...
    def __mul__(self, other): ...
    def __imul__(self, other): ...
    def __rmul__(self, other): ...
    def __matmul__(self, other): ...
    def __imatmul__(self, other): ...
    def __rmatmul__(self, other): ...
    def __div__(self, other): ...
    def __idiv__(self, other): ...
    def __rdiv__(self, other): ...
    def __truediv__(self, other): ...
    def __itruediv__(self, other): ...
    def __rtruediv__(self, other): ...
    def __pow__(self, exponent): ...
    def __ipow__(self, other): ...
    def __rpow__(self, other): ...
    @property
    def conjugate(self) -> 'Quaternion': ...
    @property
    def inverse(self) -> 'Quaternion': ...
    @property
    def norm(self) -> float: ...
    @property
    def magnitude(self) -> float: ...
    @property
    def normalised(self) -> 'Quaternion': ...
    @property
    def polar_unit_vector(self): ...
    @property
    def polar_angle(self) -> float: ...
    @property
    def polar_decomposition(self): ...
    @property
    def unit(self) -> 'Quaternion': ...
    def is_unit(self, tolerance: float = ...): ...
    def rotate(self, vector): ...
    @classmethod
    def exp(cls, q: 'Quaternion') -> 'Quaternion': ...
    @classmethod
    def log(cls, q: 'Quaternion') -> 'Quaternion': ...
    @classmethod
    def exp_map(cls, q, eta): ...
    @classmethod
    def sym_exp_map(cls, q, eta): ...
    @classmethod
    def log_map(cls, q, p): ...
    @classmethod
    def sym_log_map(cls, q, p): ...
    @classmethod
    def absolute_distance(cls, q0: 'Quaternion', q1: 'Quaternion'): ...
    @classmethod
    def distance(cls, q0: 'Quaternion', q1: 'Quaternion') -> float: ...
    @classmethod
    def sym_distance(cls, q0: 'Quaternion', q1: 'Quaternion') -> float: ...
    @classmethod
    def slerp(cls, q0: 'Quaternion', q1: 'Quaternion', amount: float = ...) -> 'Quaternion': ...
    @classmethod
    def intermediates(cls, q0: 'Quaternion', q1: 'Quaternion', n: int, include_endpoints: bool = ...) -> Generator['Quaternion', None, None]: ...
    def derivative(self, rate) -> 'Quaternion': ...
    def integrate(self, rate, timestep) -> None: ...
    @property
    def rotation_matrix(self) -> np.ndarray: ...
    @property
    def transformation_matrix(self) -> np.ndarray: ...
    @property
    def yaw_pitch_roll(self) -> Tuple[float, float, float]: ...
    def get_axis(self, undefined=...): ...
    @property
    def axis(self): ...
    @property
    def angle(self) -> float: ...
    @property
    def degrees(self) -> float: ...
    @property
    def radians(self) -> float: ...
    @property
    def scalar(self) -> float: ...
    @property
    def vector(self): ...
    @property
    def real(self) -> float: ...
    @property
    def imaginary(self): ...
    @property
    def w(self) -> float: ...
    @property
    def x(self) -> float: ...
    @property
    def y(self) -> float: ...
    @property
    def z(self) -> float: ...
    @property
    def elements(self): ...
    def __getitem__(self, index: int) -> float: ...
    def __setitem__(self, index: int, value: float) -> None: ...
    def __copy__(self) -> 'Quaternion': ...
    def __deepcopy__(self, memo) -> 'Quaternion': ...
    @staticmethod
    def to_degrees(angle_rad: Optional[float]) -> Optional[float]: ...
    @staticmethod
    def to_radians(angle_deg: Optional[float]) -> Optional[float]: ...



"""The top-level Processor class for marker detection and annotation."""
import logging
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from numpy.typing import NDArray
from pyapriltags import Detector

from .frame_sources import FrameSource
from .marker import Marker
from .utils import Frame, normalise_marker_text

LOGGER = logging.getLogger(__name__)


class Processor:
    """A pyaprilTags wrapper for fiducial detection from frame sources."""

    capture_filter: Callable[[NDArray], NDArray]
    marker_filter: Callable[[List[Marker]], List[Marker]]
    detection_hook: Callable[[Frame, List[Marker]], None]

    def __init__(
        self,
        frame_source: Optional[FrameSource] = None,
        calibration: Optional[Tuple[float, float, float, float]] = None,
        tag_sizes: Union[float, Dict[int, float], None] = None,
        *,
        tag_family: str = 'tag36h11',
        threads: int = 4,
        quad_decimate: float = 2,
        aruco_orientation: bool = True,
        name: str = "Camera",
        vidpid: str = "",
    ) -> None:
        self.capture_filter = lambda frame: frame
        self.marker_filter = lambda markers: markers
        self.detection_hook = lambda frame, markers: None

        self._aruco_orientation = aruco_orientation
        self.calibration = calibration
        self.name = name
        self.vidpid = vidpid

        if frame_source is None:
            frame_source = FrameSource()
        self._frame_source = frame_source

        if tag_sizes is None:
            tag_sizes = {}
        self.tag_sizes = tag_sizes

        self.detector = Detector(
            families=tag_family,
            nthreads=threads,
            quad_decimate=quad_decimate,
        )

    def _capture(self, fresh: bool = True) -> Frame:
        """Get another frame from the underlying camera."""
        colour_frame = self._frame_source.read(fresh)

        # hook to allow modification of the captured frame
        colour_frame = self.capture_filter(colour_frame)

        return Frame.from_colour_frame(colour_frame)

    def _detect(self, frame: Frame) -> List[Marker]:
        """Locate fiducial markers in frame using pyapriltags."""
        if self.calibration is None:
            detections = self.detector.detect(frame.grey_frame)
        else:
            detections = self.detector.detect(
                frame.grey_frame,
                estimate_tag_pose=True,
                camera_params=self.calibration,
                tag_size=self.tag_sizes,
            )

        markers: List[Marker] = []
        for detection in detections:
            markers.append(Marker(
                detection,
                aruco_orientation=self._aruco_orientation,
            ))

        # hook to filter and modify markers
        markers = self.marker_filter(markers)

        # hook to extract a frame and its markers
        self.detection_hook(frame, markers)

        return markers

    def _annotate(
        self,
        frame: Frame,
        markers: List[Marker],
        line_thickness: int = 2,
        text_scale: float = 1,
    ) -> Frame:
        """
        Annotate marker borders and ids onto frame.

        The annotation is in-place.
        """
        for marker in markers:
            integer_corners = np.array(marker.pixel_corners, dtype=np.int32)
            marker_id = f"id={marker.id}"

            for frame_type in frame:
                cv2.polylines(
                    frame_type,
                    [integer_corners],
                    isClosed=True,
                    color=(0, 255, 0),  # Green (BGR)
                    thickness=line_thickness,
                )

                # Add square at marker origin corner
                if self._aruco_orientation:
                    origin_square = np.array([
                        # index 1 is the top-left corner in aruco
                        integer_corners[1] + np.array([3, 3]),
                        integer_corners[1] + np.array([3, -3]),
                        integer_corners[1] + np.array([-3, -3]),
                        integer_corners[1] + np.array([-3, 3]),
                    ], dtype=np.int32)
                else:
                    origin_square = np.array([
                        # index -1 is the top-left corner in apriltag
                        integer_corners[-1] + np.array([3, 3]),
                        integer_corners[-1] + np.array([3, -3]),
                        integer_corners[-1] + np.array([-3, -3]),
                        integer_corners[-1] + np.array([-3, 3]),
                    ], dtype=np.int32)

                cv2.polylines(
                    frame_type,
                    [origin_square],
                    isClosed=True,
                    color=(0, 0, 255),  # red
                    thickness=line_thickness,
                )

                marker_text_scale = text_scale * normalise_marker_text(marker)

                # Approximately center the text
                text_origin = np.array(marker.pixel_centre, dtype=np.int32)
                text_origin += np.array(
                    [-40 * marker_text_scale, 10 * marker_text_scale], dtype=np.int32)

                cv2.putText(
                    frame_type,
                    marker_id,
                    text_origin,
                    cv2.FONT_HERSHEY_DUPLEX,
                    marker_text_scale,
                    color=(255, 191, 0),  # deep sky blue
                    thickness=2,
                )

        return frame

    def _save(self, frame: Frame, name: Union[str, Path], colour: bool = True) -> None:
        """Save a frame to a file, selectable between colour and grayscale."""
        if colour:
            output_frame = frame.colour_frame
        else:
            output_frame = frame.grey_frame

        path = Path(name)
        if not path.suffix:
            LOGGER.warning("No file extension given, defaulting to jpg")
            path = path.with_suffix(".jpg")

        cv2.imwrite(str(path), output_frame)

    def capture(self) -> NDArray:
        """Get the raw image data from the camera."""
        return self._capture().colour_frame

    def see(self, *, frame: Optional[NDArray] = None) -> List[Marker]:
        """Get markers that the camera can see."""
        if frame is None:
            frames = self._capture()
        else:
            frames = Frame.from_colour_frame(frame)
        return self._detect(frames)

    def see_ids(self, *, frame: Optional[NDArray] = None) -> List[int]:
        """Get a list of visible marker IDs."""
        if frame is None:
            frames = self._capture()
        else:
            frames = Frame.from_colour_frame(frame)
        markers = self._detect(frames)
        return [marker.id for marker in markers]

    def save(self, name: Union[str, Path], *, frame: Optional[NDArray] = None) -> None:
        """Save an annotated image to a file."""
        if frame is None:
            frames = self._capture()
        else:
            frames = Frame.from_colour_frame(frame)
        markers = self._detect(frames)
        frames = self._annotate(
            frames,
            markers,
        )
        self._save(frames, name)

    def close(self) -> None:
        """Close the underlying capture device."""
        self._frame_source.close()



"""OS-specific and generic methods for detecting attached USB cameras."""
import json
import logging
import re
import subprocess
import sys
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Tuple

import cv2

LOGGER = logging.getLogger(__name__)


class CameraIdentifier(NamedTuple):
    """A tuple to store information of a discovered camera."""

    index: int  # type: ignore[assignment]
    name: str
    vidpid: str


class CalibratedCamera(NamedTuple):
    """A tuple to store camera and calibration information of a discovered camera."""

    index: int  # type: ignore[assignment]
    name: str
    vidpid: str
    calibration: Optional[Path] = None


def find_cameras(
    calibration_locations: List[str],
    include_uncalibrated: bool = False,
) -> List[CalibratedCamera]:
    """Find the available cameras using OS-specific discovery where available."""
    platform = sys.platform

    calibration_map = generate_calibration_file_map(calibration_locations)

    if platform.startswith("linux"):
        cameras = linux_discovery()
    elif platform.startswith("darwin"):
        cameras = mac_discovery()
    elif platform == "win32":
        valid_cameras = windows_discovery(calibration_map, include_uncalibrated)
        return valid_cameras
    else:
        cameras = default_discovery()

    valid_cameras = match_calibrations(cameras, calibration_map, include_uncalibrated)

    return valid_cameras


def generate_calibration_file_map(calibration_locations: List[str]) -> Dict[str, Path]:
    """
    Generate map of USB VID & PID to calibration file.

    Uses vidpid tag in XML files.
    """
    calibration_map: Dict[str, Path] = {}
    # iterate calibration locations backword so that so that earlier locations
    # in the list get higher precedence
    for location in reversed(calibration_locations):
        for calibration_file in Path(location).glob('*.xml'):
            storage = cv2.FileStorage(str(calibration_file), cv2.FILE_STORAGE_READ)

            node = storage.getNode('vidpid')
            if node.isSeq():
                pidvids = [node.at(i).string() for i in range(node.size())]
            elif node.isString():
                pidvids = [node.string()]
            else:
                # This file lacks any vidpids
                continue

            for pidvid in pidvids:
                calibration_map[pidvid] = calibration_file.absolute()
    LOGGER.debug(f"Calibrations found for: {list(calibration_map.keys())}")

    return calibration_map


def match_calibrations(
    cameras: List[CameraIdentifier],
    calibration_map: Dict[str, Path],
    include_uncalibrated: bool,
) -> List[CalibratedCamera]:
    """Filter found cameras to those that matching calibration files' USB VID & PID."""
    calibrated_cameras: List[CalibratedCamera] = []
    uncalibrated_cameras: List[CalibratedCamera] = []

    for camera in cameras:
        try:
            calibrated_cameras.append(CalibratedCamera(
                index=camera.index,
                name=camera.name,
                vidpid=camera.vidpid,
                calibration=calibration_map[camera.vidpid],
            ))
            LOGGER.debug(
                f"Found calibration for {camera.name} in {calibration_map[camera.vidpid]}")
        except KeyError:
            LOGGER.debug(f"No calibration found for for {camera.name}")
            if include_uncalibrated is True:
                uncalibrated_cameras.append(CalibratedCamera(
                    index=camera.index,
                    name=camera.name,
                    vidpid=camera.vidpid,
                ))

    return calibrated_cameras + uncalibrated_cameras


def usable_present_devices(calibration_map: Dict[str, Path]) -> List[Tuple[str, Path, str]]:
    """Use PyUSB to detect any supported USB VID/PID combinations connected to the system."""
    import libusb_package
    try:
        usb_devices = libusb_package.find(find_all=True)
    except ValueError:
        LOGGER.warning("libusb_package failed to find a libusb backend.")
        return []

    usable_devices: List[Tuple[str, Path, str]] = []

    for dev in usb_devices:
        vidpid = f"{dev.idVendor:04x}:{dev.idProduct:04x}"
        calibration = calibration_map.get(vidpid)
        if calibration is not None:
            usable_devices.append((vidpid, calibration, calibration.stem))

    LOGGER.debug(f"Found calibration for the devices: {[dev[0] for dev in usable_devices]}")
    return usable_devices


def linux_discovery() -> List[CameraIdentifier]:
    """
    Discovery method for Linux using Video4Linux.

    This matches camera indexes to their USB VID & PID and omits indexes
    that are not the actual camera.
    """
    # Get a list of all potential cameras
    potential_cameras = []
    for index in range(10):
        if Path(f'/dev/video{index}').exists():
            potential_cameras.append(index)

    # Filter the potential cameras to ones that work
    valid_cameras = []
    for index in potential_cameras:
        capture = cv2.VideoCapture(f'/dev/video{index}')
        if capture.isOpened():
            valid_cameras.append(index)
        capture.release()

    # Match pid:vid to the path
    cameras = []
    for index in valid_cameras:
        name = Path(
                f'/sys/class/video4linux/video{index}/name',
            ).read_text().strip()

        uevent_file = Path(f'/sys/class/video4linux/video{index}/device/uevent').read_text()
        m = re.search(r'PRODUCT=([0-9a-f]{1,4})\/([0-9a-f]{1,4})', uevent_file)

        if m is None:
            continue

        vid = int(m.groups()[0], 16)
        pid = int(m.groups()[1], 16)

        vidpid = f'{vid:04x}:{pid:04x}'

        LOGGER.debug(f"Found camera at index {index}: {name}")
        cameras.append(CameraIdentifier(
            index=index,
            name=name,
            vidpid=vidpid,
        ))

    return cameras


def mac_discovery() -> List[CameraIdentifier]:
    """
    Discovery method for MacOS using system_profiler.

    This matches camera indexes to their USB VID & PID as long as cameras are
    not attached after cv2 has been imported.
    """
    camera_info = json.loads(
        subprocess.check_output(['system_profiler', '-json', 'SPCameraDataType']),
    )
    camera_list = camera_info["SPCameraDataType"]
    cameras = []

    for index, camera in enumerate(camera_list):
        try:
            name = camera["_name"]
            camera_data = camera['spcamera_model-id']

            m = re.search(r'VendorID_([0-9]{1,5}) ProductID_([0-9]{1,5})', camera_data)

            if m is None:
                # Facetime cameras have no PID or VID
                vidpid = camera_data
            else:
                vid = int(m.groups()[0], 10)
                pid = int(m.groups()[1], 10)

                vidpid = f'{vid:04x}:{pid:04x}'

            LOGGER.debug(f"Found camera at index {index}: {name}")
            cameras.append(CameraIdentifier(index=index, name=name, vidpid=vidpid))
        except KeyError:
            LOGGER.warning(f"Camera {index} had missing fields: {camera}")

    return cameras


def windows_discovery(
    calibration_map: Dict[str, Path],
    include_uncalibrated: bool,
) -> List[CalibratedCamera]:
    """
    Discovery method for Windows using PyUSB and the fallback discovery method.

    This cannot identify which index corresponds to the found USB VID & PID
    so is unable to provide a useful output when more than one USB camera with
    calibration is connected.
    The returned camera is a combination of the found USB VID & PID and the
    first openable camera index.
    """
    found_cameras = default_discovery()
    if include_uncalibrated:
        # We lack the information to match which camera is which so treat them
        # all as uncalibrated
        LOGGER.debug("Assuming all cameras are uncalibrated")
        return [
            CalibratedCamera(
                index=camera.index,
                name=camera.name,
                vidpid=camera.vidpid,
            ) for camera in found_cameras
        ]

    if len(found_cameras) == 0:
        return []

    selected_camera = found_cameras[0]
    LOGGER.debug(f"Selecting camera index {selected_camera.index}")

    usable_cameras = usable_present_devices(calibration_map)
    if len(usable_cameras) > 1:
        raise RuntimeError(
            "Naive calibration selection is only supported when a single compatible "
            "device is connected")
    elif len(usable_cameras) == 1:
        usable_camera = usable_cameras[0]
        return [CalibratedCamera(
            index=selected_camera.index,
            name=usable_camera[2],
            vidpid=usable_camera[0],
            calibration=usable_camera[1],
        )]
    else:
        return []


def default_discovery() -> List[CameraIdentifier]:
    """
    A fallback discovery method for when there is not an OS specific one available.

    This cannot identify the USB VID & PID of the camera and only provides
    information on the openable indexes.
    """
    found_cameras = []
    for camera_id in range(8):
        capture = cv2.VideoCapture(camera_id)
        if capture.isOpened():
            LOGGER.debug(f"Found camera at index {camera_id}")
            found_cameras.append(CameraIdentifier(
                index=camera_id,
                name=str(camera_id),
                vidpid='',
            ))
        capture.release()

    return found_cameras



"""
A charuco calibration script.

From: https://gist.github.com/naoki-mizuno/d25cbc3c59228291cabe50529d70894c
"""
import argparse
import logging
import sys

import cv2
import numpy as np

LOGGER = logging.getLogger(__name__)


def read_chessboards(frames, aruco_dict, board):
    """Charuco base pose estimation."""
    all_corners = []
    all_ids = []

    for frame in frames:
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        corners, ids, rejectedImgPoints = cv2.aruco.detectMarkers(gray, aruco_dict)

        if len(corners) > 0:
            ret, c_corners, c_ids = cv2.aruco.interpolateCornersCharuco(
                corners, ids, gray, board)
            # ret is the number of detected corners
            if ret > 0:
                all_corners.append(c_corners)
                all_ids.append(c_ids)
        else:
            LOGGER.error('Failed!')

    imsize = gray.shape
    return all_corners, all_ids, imsize


def capture_camera(cap, num=1, mirror=False, size=None):
    """Capture frames to be used for calibration."""
    frames = []
    LOGGER.info("Press space to capture frame.")

    while True:
        ret, frame = cap.read()

        if mirror is True:
            frame = cv2.flip(frame, 1)

        if size is not None and len(size) == 2:
            frame = cv2.resize(frame, size)

        cv2.imshow('Frame', frame)

        k = cv2.waitKey(1)
        if k == 27:  # Esc
            break
        elif k == 10 or k == 32:  # Enter or Space
            frames.append(frame)
            LOGGER.info('Frame captured!')
            if len(frames) == num:
                break

    return frames


def draw_axis(frame, camera_matrix, dist_coeff, aruco_dict, board):
    """Annotate charuco marker target using calibration parameters."""
    corners, ids, rejected_points = cv2.aruco.detectMarkers(frame, aruco_dict)
    ret, c_corners, c_ids = cv2.aruco.interpolateCornersCharuco(
        corners, ids, frame, board)
    ret, p_rvec, p_tvec = cv2.aruco.estimatePoseCharucoBoard(
        c_corners, c_ids, board, camera_matrix, dist_coeff, None, None)
    cv2.drawFrameAxes(
        frame, camera_matrix, dist_coeff,  p_rvec, p_tvec, 0.1)
    # cv2.aruco.drawDetectedCornersCharuco(frame, c_corners, c_ids)
    # cv2.aruco.drawDetectedMarkers(frame, corners, ids)
    # cv2.aruco.drawDetectedMarkers(frame, rejected_points, borderColor=(100, 0, 240))
    LOGGER.info('Translation : {0}'.format(p_tvec))
    LOGGER.info('Rotation    : {0}'.format(p_rvec))
    LOGGER.info('Distance from camera: {0} m'.format(np.linalg.norm(p_tvec)))

    return frame


def main(args: argparse.Namespace):
    """Charuco calibration."""
    try:
        import cv2.aruco  # type: ignore
    except ImportError:
        LOGGER.critical("Calibration requires the opencv-contrib-python package")
        sys.exit(1)

    aruco_dict = cv2.aruco.Dictionary_get(cv2.aruco.DICT_4X4_50)
    board = cv2.aruco.CharucoBoard_create(8, 6, 0.03, 0.023, aruco_dict)

    dev_num = args.camera
    video_dev = cv2.VideoCapture(dev_num)

    if args.resolution:
        video_dev.set(cv2.CAP_PROP_FRAME_WIDTH, args.resolution[0])
        video_dev.set(cv2.CAP_PROP_FRAME_HEIGHT, args.resolution[1])

    frames = capture_camera(video_dev, args.frame_count)
    if len(frames) == 0:
        LOGGER.error('No frame captured')
        sys.exit(1)
    all_corners, all_ids, imsize = read_chessboards(frames, aruco_dict, board)
    ret, camera_matrix, dist_coeff, rvec, tvec = cv2.aruco.calibrateCameraCharuco(
        all_corners, all_ids, board, imsize, None, None,
    )

    LOGGER.info(f"> Camera matrix\n{camera_matrix}")
    LOGGER.info(f"> Distortion coefficients\n{dist_coeff}")

    height = int(video_dev.get(cv2.CAP_PROP_FRAME_WIDTH))
    width = int(video_dev.get(cv2.CAP_PROP_FRAME_HEIGHT))
    LOGGER.info(f"> Resolution\n{height}x{width}")

    LOGGER.info("Capture frame to test calibration.")
    test_frame = capture_camera(video_dev, 1)[0]
    test_frame = draw_axis(test_frame, camera_matrix, dist_coeff, aruco_dict, board)

    cv2.imshow('Frame', test_frame)
    cv2.waitKey(0)

    video_dev.release()


def create_subparser(subparsers: argparse._SubParsersAction):
    """Calibrate command parser."""
    parser = subparsers.add_parser(
        "calibrate",
        help="Generate camera calibration",
        description="""
        Generate camera calibration using a ChArUco board.

        Generate a board with:
        https://calib.io/pages/camera-calibration-pattern-generator
        260x200mm, 6x8 squares, 30mm checkers.
        """)
    parser.add_argument('camera', type=int, help="The camera index")
    parser.add_argument(
        '--resolution', type=int, nargs=2, default=None,
        help="Force camera resolution")
    parser.add_argument(
        '-n', '--frame_count', type=int, default=15,
        help="Number of frames to use for calbration.")

    parser.set_defaults(func=main)



"""An AprilTags wrapper with camera discovery and axis conversion."""
from ._version import __version__
from .calibrations import calibrations
from .detect_cameras import find_cameras
from .frame_sources import FrameSource, USBCamera
from .marker import (CartesianCoordinates, Marker, Orientation,
                     PixelCoordinates, SphericalCoordinate)
from .vision import Processor

__all__ = [
    '__version__',
    'CartesianCoordinates',
    'FrameSource',
    'Marker',
    'Orientation',
    'PixelCoordinates',
    'Processor',
    'SphericalCoordinate',
    'USBCamera',
    'calibrations',
    'find_cameras',
]



"""The available frame sources for the Processor class."""
import logging
from pathlib import Path
from sys import platform
from typing import List, Optional, Tuple, Union

import cv2
from numpy.typing import NDArray

LOGGER = logging.getLogger(__name__)


class FrameSource:
    """
    A base frame source.

    Allows april_vision.Processor to be created prior to frames being available.
    """

    def read(self, fresh: bool = True) -> NDArray:
        """
        The method for getting a new frame.

        :param fresh: Whether to flush the device's buffer before capturing
        the frame, unused in the base class.
        """
        raise NotImplementedError("This frame source does not produce any frames.")

    def close(self) -> None:
        """Default method for closing the underlying capture device."""
        pass


class USBCamera(FrameSource):
    """A USB attached camera."""

    def __init__(
        self,
        index: int,
        resolution: Tuple[int, int] = (0, 0),
        camera_parameters: Optional[List[Tuple[int, int]]] = None,
        calibration: Optional[Tuple[float, float, float, float]] = None,
        vidpid: str = "",
    ) -> None:
        """
        Create a USB attached camera frame source.

        :param index: The camera's opencv index.
        :param resolution: Resolution to set the camera to.
        :param camera_parameters: Additional opencv parameters to apply to the camera.
        :param calibration: Optional, the intrinsic parameters of the camera.
        :param vidpid: The vendor/product string for the camera.
        """
        self._camera = cv2.VideoCapture(index)

        if resolution != (0, 0):
            try:
                self._set_resolution(resolution)
            except AssertionError as e:
                LOGGER.warning(f"Failed to set resolution: {e}")

        if camera_parameters is None:
            camera_parameters = []

        # Set buffer length to 1
        camera_parameters.append((cv2.CAP_PROP_BUFFERSIZE, 1))

        for parameter, value in camera_parameters:
            try:
                self._set_camera_property(parameter, value)
            except AssertionError as e:
                LOGGER.warning(f"Failed to set property: {e}")

        # Maximise the framerate on Linux
        # NOTE optimisation is disabled currently due to
        # 'Corrupt JPEG data: 5 extraneous bytes' seen on raspberry pi 4's
        # self._optimise_camera(vidpid)

        self._buffer_length = int(self._camera.get(cv2.CAP_PROP_BUFFERSIZE))

        # Take and discard a camera capture
        _ = self._capture_single_frame()

        # Save the camera intrinsic parameters
        self.calibration = calibration

    @classmethod
    def from_calibration_file(
        cls,
        index: int,
        calibration_file: Union[str, Path, None],
        vidpid: str = "",
    ) -> 'USBCamera':
        """Instantiate camera using calibration data from opencv XML calibration file."""
        if calibration_file is not None:
            calibration_file = Path(calibration_file)
        else:
            return cls(index)

        if not calibration_file.exists():
            LOGGER.warning(f"Calibrations not found: {calibration_file}")
            return cls(index)

        storage = cv2.FileStorage(str(calibration_file), cv2.FILE_STORAGE_READ)
        resolution_node = storage.getNode("cameraResolution")
        camera_matrix = storage.getNode("cameraMatrix").mat()
        fx, fy = camera_matrix[0, 0], camera_matrix[1, 1]
        cx, cy = camera_matrix[0, 2], camera_matrix[1, 2]
        return cls(
            index,
            resolution=(
                int(resolution_node.at(0).real()),
                int(resolution_node.at(1).real()),
            ),
            calibration=(fx, fy, cx, cy),
            vidpid=vidpid,
        )

    def _set_camera_property(self, property: int, value: int) -> None:  # noqa: A002
        """Set an opencv property to a value and assert that it changed."""
        self._camera.set(property, value)
        actual = self._camera.get(property)

        assert actual == value, (f"Failed to set property '{property}', "
                                 f"expected {value} got {actual}")

    def _set_resolution(self, resolution: Tuple[int, int]) -> None:
        """Set the camera resolution and assert that it changed."""
        width, height = resolution
        self._camera.set(cv2.CAP_PROP_FRAME_WIDTH, width)
        self._camera.set(cv2.CAP_PROP_FRAME_HEIGHT, height)

        actual = self._get_resolution()
        assert actual == resolution, ("Failed to set resolution expected "
                                      f"{resolution} got {actual}")

    def _get_resolution(self) -> Tuple[int, int]:
        """Get the camera resolution active in opencv."""
        return (
            int(self._camera.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(self._camera.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )

    def _optimise_camera(self, vidpid: str):
        """Tweak the camera's image type and framerate to achieve the minimum frame time."""
        verified_vidpid = {'046d:0825', '046d:0807'}
        if not platform.startswith("linux"):
            # All current optimisations are for linux
            return

        # These may not improve frame time on all cameras
        if vidpid not in verified_vidpid:
            return

        camera_parameters = [
            (cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*'MJPG')),
            (cv2.CAP_PROP_FPS, 30),
            (cv2.CAP_PROP_BUFFERSIZE, 2),
        ]

        LOGGER.debug("Optimising camera")

        for parameter, value in camera_parameters:
            try:
                self._set_camera_property(parameter, value)
            except AssertionError as e:
                LOGGER.warning(f"Failed to set property: {e}")

    def _capture_single_frame(self) -> NDArray:
        """Read a single frame from the camera's buffer."""
        ret, colour_frame = self._camera.read()
        if not ret:
            raise IOError("Failed to get frame from camera")
        return colour_frame

    def read(self, fresh: bool = True) -> NDArray:
        """
        Get another frame from the camera.

        :param fresh: Whether to flush the device's buffer before capturing
        the frame.
        """
        if fresh is True:
            for _ in range(self._buffer_length):
                # Discard a frame to remove the old frame in the buffer
                _ = self._capture_single_frame()

        return self._capture_single_frame()

    def close(self) -> None:
        """Close the underlying capture device."""
        self._camera.release()


class VideoSource(FrameSource):
    """Return frames from a video."""

    def __init__(self, filepath: Union[str, Path]) -> None:
        """
        Read a video from file.

        :param filepath: The path to the video to load.
        """
        self._video = cv2.VideoCapture(str(filepath))

    def read(self, fresh: bool = True) -> NDArray:
        """
        Get the next frame from the video file.

        :param fresh: Unused.
        """
        ret, colour_frame = self._video.read()
        if not ret:
            raise IOError("Failed to get frame from video")
        return colour_frame

    def close(self) -> None:
        """Close the video file."""
        self._video.release()


class ImageSource(FrameSource):
    """Return a single image repeatedly."""

    def __init__(self, filepath: Union[str, Path]) -> None:
        """
        Load an image from file.

        :param filepath: The path to the image to load.
        """
        self._frame = cv2.imread(str(filepath))

    def read(self, fresh: bool = True) -> NDArray:
        """Return the stored frame."""
        return self._frame



import array
from typing import Callable, Iterable, Iterator, Optional, Tuple, Union

from usb.backend import IBackend

class _DescriptorInfo(str): ...

class USBError(IOError): ...
class USBTimeoutError(USBError): ...
class NoBackendError(ValueError): ...

class Endpoint: ...

class Interface: ...

class Configuration: ...

class Device:
    bLength: int
    bDescriptorType: int
    bcdUSB: int
    bDeviceClass: int
    bDeviceSubClass: int
    bDeviceProtocol: int
    bMaxPacketSize0: int
    idVendor: int
    idProduct: int
    bcdDevice: int
    iManufacturer: int
    iProduct: int
    iSerialNumber: int
    bNumConfigurations: int
    address: Optional[int]
    bus: Optional[int]
    port_number: Optional[int]
    port_numbers: Tuple[int, ...]
    speed: Optional[int]

    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def configurations(self) -> Tuple['Device']: ...
    def __init__(self, dev, backend: IBackend) -> None: ...
    @property
    def langids(self) -> Tuple[int]: ...
    @property
    def serial_number(self) -> str: ...
    @property
    def product(self) -> str: ...
    @property
    def parent(self) -> Optional['Device']: ...
    @property
    def manufacturer(self) -> str: ...
    @property
    def backend(self) -> IBackend: ...
    def set_configuration(self, configuration: Optional[int] = None) -> None: ...
    def get_active_configuration(self) -> Configuration: ...
    def set_interface_altsetting(self, interface: Optional[int] = None, alternate_setting: Optional[int] = None) -> None: ...
    def clear_halt(self, ep: Union[int, Endpoint]) -> None: ...
    def reset(self) -> None: ...
    def write(self, endpoint: int, data: Union[array.array, str, bytes, None], timeout: Optional[int] = None): ...
    def read(self, endpoint: int, size_or_buffer: Union[int, array.array], timeout: Optional[int] = None): ...
    def ctrl_transfer(self, bmRequestType: int, bRequest: int, wValue: int = 0, wIndex: int = 0, data_or_wLength:  Union[array.array, str, bytes, None, int] = None, timeout = None): ...
    def is_kernel_driver_active(self, interface: int) -> bool: ...
    def detach_kernel_driver(self, interface: int) -> None: ...
    def attach_kernel_driver(self, interface: int) -> None: ...
    def __iter__(self) -> Iterator[Configuration]: ...
    def __getitem__(self, index: int) -> Configuration: ...
    @property
    def default_timeout(self) -> float: ...
    def finalize(self) -> None: ...


def find(find_all: bool = ..., backend: IBackend = ..., custom_match: Optional[Callable[[Device], bool]] = ..., **args) -> Union[Device, Iterable[Device]]: ...
def show_devices(verbose: bool = ..., **kwargs) -> _DescriptorInfo: ...



"""
Annotate an image file.

Takes an input image, adds annotation and saves the image.
"""
import argparse
import logging
from pathlib import Path

import cv2

from ..frame_sources import ImageSource
from ..marker import MarkerType
from ..utils import annotate_text, load_calibration, normalise_marker_text
from ..vision import Processor

LOGGER = logging.getLogger(__name__)


def main(args: argparse.Namespace):
    """Annotate an image file using the provided args."""
    input_file = Path(args.input_file)
    if not input_file.exists():
        LOGGER.fatal("Input file does not exist.")
        return

    calibration = None
    if args.calibration:
        _, calibration = load_calibration(args.calibration)

    source = ImageSource(args.input_file)
    processer = Processor(
        source,
        tag_family=args.tag_family,
        quad_decimate=args.quad_decimate,
        tag_sizes=float(args.tag_size) / 1000,
        calibration=calibration,
    )

    frame = processer._capture()
    markers = processer._detect(frame)
    LOGGER.info(f"Found {len(markers)} markers.")
    frame = processer._annotate(frame, markers)

    if args.calibration:
        try:
            for marker in markers:
                # Check we have pose data
                _ = marker.cartesian

                text_scale = normalise_marker_text(marker)

                loc = (
                    int(marker.pixel_centre[0] - 80 * text_scale),
                    int(marker.pixel_centre[1] + 40 * text_scale),
                )
                frame = annotate_text(
                    frame, f"dist={marker.distance}mm", loc,
                    text_scale=0.8 * text_scale,
                    text_colour=(255, 191, 0),  # deep sky blue
                )
        except RuntimeError:
            pass

    # save frame
    cv2.imwrite(args.output_file, frame.colour_frame)

    if args.preview:
        cv2.imshow('image', frame.colour_frame)
        LOGGER.info("Press any key to close window.")
        _ = cv2.waitKey(0)
        cv2.destroyAllWindows()


def create_subparser(subparsers: argparse._SubParsersAction):
    """Annotate_image command parser."""
    parser = subparsers.add_parser(
        "annotate_image",
        description="Annotate an image file with its markers",
        help="Annotate an image file with its markers",
    )

    parser.add_argument("input_file", type=str, help="The image to process.")
    parser.add_argument("output_file", type=str, help="The filepath to save the output to.")

    parser.add_argument('--preview', action='store_true', help="Display the annotated image.")

    parser.add_argument(
        '--tag_family', default=MarkerType.APRILTAG_36H11.value,
        choices=[marker.value for marker in MarkerType],
        help="Set the marker family to detect, defaults to 'tag36h11'")
    parser.add_argument(
        '--quad_decimate', type=float, default=2,
        help="Set the level of decimation used in the detection stage")

    parser.add_argument(
        '--calibration', type=Path, default=None, help="Calbration XML file to use.")
    parser.add_argument(
        '--tag_size', type=int, default=0, help="The size of markers in millimeters")

    parser.set_defaults(func=main)



"""
Provide a folder of images and iterate over the images doing marker detection.

Give and output of the results.
Can provide a list of the actual answers and get a result of the error.
"""
import argparse

# TODO impliment
# marker_benchmark
#     folder of test images
# 	provide list of actual answers to check against
#     give performance result


def main(args: argparse.Namespace):
    """Iterate over images doing marker detection."""
    raise NotImplementedError("Marker benchmark is not implemented.")


def create_subparser(subparsers: argparse._SubParsersAction):
    """Marker_benchmark command parser."""
    parser = subparsers.add_parser("marker_benchmark")

    parser.set_defaults(func=main)



"""j5 integration for april_vision."""
import base64
import logging
import os
from pathlib import Path
from threading import Thread
from typing import Callable, Dict, Iterable, List, Optional, Set, Type, Union

import cv2
import numpy as np
from j5.backends import Backend
from j5.boards import Board
from j5.components.component import Component
from numpy.typing import NDArray

from .._version import __version__
from ..detect_cameras import CalibratedCamera, find_cameras
from ..frame_sources import USBCamera
from ..marker import Marker, MarkerType
from ..utils import Frame
from ..vision import Processor

LOGGER = logging.getLogger(__name__)


class AprilCameraBoard(Board):
    """
    Virtual Camera Board for detecting fiducial markers.

    Additionally, it will do pose estimation, along with some calibration
    in order to determine the spatial positon and orientation of the markers
    that it has detected.
    """

    name: str = "AprilTag Camera Board"
    _backend: 'AprilTagHardwareBackend'

    def __init__(self, serial: str, backend: 'AprilTagHardwareBackend'):
        self._serial = serial
        self._backend = backend

    @property
    def serial_number(self) -> str:
        """Get the serial number."""
        return self._serial

    @property
    def firmware_version(self) -> Optional[str]:
        """Get the firmware version of the board."""
        return self._backend.firmware_version

    def make_safe(self) -> None:
        """
        Close the camera.

        The camera will no longer work after this method is called.
        """
        self._backend.close_camera()

    @staticmethod
    def supported_components() -> Set[Type[Component]]:
        """List the types of components supported by this board."""
        return set()

    # Proxy methods from MarkerCamera object
    def see(self, *, eager: bool = True, frame: Optional[NDArray] = None) -> List[Marker]:
        """
        Capture an image and identify fiducial markers.

        :param eager: Process the pose estimations of markers immediately,
            currently unused.
        :returns: list of markers that the camera could see.
        """
        return self._backend.see(frame=frame)

    def see_ids(self, *, frame: Optional[NDArray] = None) -> List[int]:
        """
        Capture an image and identify fiducial markers.

        This method returns just the marker IDs that are visible.
        :returns: A list of IDs for the markers that were visible.
        """
        return self._backend.see_ids(frame=frame)

    def capture(self) -> NDArray:
        """
        Get the raw image data from the camera.

        :returns: Camera pixel data
        """
        return self._backend.capture_frame()

    def save(self, path: Union[Path, str], *, frame: Optional[NDArray] = None) -> None:
        """Save an annotated image to a path."""
        self._backend.save_annotated_image(path, frame=frame)


class AprilTagHardwareBackend(Backend):
    """An April Vision Hardware backend."""

    board = AprilCameraBoard
    marker_type = MarkerType.APRILTAG_36H11

    @classmethod
    def discover(cls) -> Set[Board]:
        """Discover boards that this backend can control."""
        return {
            AprilCameraBoard(
                f"{camera_data.name} - {camera_data.index}",
                cls(camera_data.index, camera_data=camera_data),
            )
            for camera_data in find_cameras(
                os.environ.get('OPENCV_CALIBRATIONS', '.').split(':'))
        }

    def __init__(self, camera_id: int, camera_data: CalibratedCamera) -> None:
        """Generate a backend from the camera index and calibration data."""
        camera_source = USBCamera.from_calibration_file(
            camera_id,
            calibration_file=camera_data.calibration,
            vidpid=camera_data.vidpid,
        )
        self._cam = Processor(
            camera_source,
            calibration=camera_source.calibration,
            name=camera_data.name,
            vidpid=camera_data.vidpid,
        )
        self._marker_offset = 0
        self._cam.marker_filter = self.marker_filter
        self._mqtt_publish: Optional[Callable[[str, Union[bytes, str]], None]] = None
        self._cam.detection_hook = self.annotated_frame_hook

    def see(
        self,
        *,
        frame: Optional[NDArray] = None,
    ) -> List[Marker]:
        """Get markers that the camera can see."""
        return self._cam.see(frame=frame)

    def save_annotated_image(
        self, file: Union[Path, str], *, frame: Optional[NDArray] = None,
    ) -> None:
        """Save an annotated image to a file."""
        self._cam.save(file, frame=frame)

    def see_ids(self, *, frame: Optional[NDArray] = None) -> List[int]:
        """
        Get a list of visible marker IDs.

        :returns: List of marker IDs that were visible.
        """
        return self._cam.see_ids(frame=frame)

    def capture_frame(self) -> NDArray:
        """
        Get the raw image data from the camera.

        :returns: Camera pixel data
        """
        return self._cam.capture()

    @property
    def firmware_version(self) -> Optional[str]:
        """The firmware version of the board."""
        return f"April camera v{__version__}"

    def close_camera(self) -> None:
        """Close the camera object."""
        self._cam.close()

    def set_marker_sizes(
        self,
        marker_sizes: Dict[Iterable[int], int],
        marker_offset: int = 0,
    ) -> None:
        """Set the sizes of all the markers used in the game."""
        # store marker offset to be used by the filter
        self._marker_offset = marker_offset
        # Reset previously stored sizes
        self._cam.tag_sizes = {}
        for marker_ids, marker_size in marker_sizes.items():
            # Unroll generators to give direct lookup
            for marker_id in marker_ids:
                # Convert to meters
                self._cam.tag_sizes[marker_id + marker_offset] = float(marker_size) / 1000

    def marker_filter(self, markers: List[Marker]) -> List[Marker]:
        """Apply marker offset and remove markers that are not in the game."""
        filtered_markers: List[Marker] = []
        if not isinstance(self._cam.tag_sizes, dict):
            for marker in markers:
                marker._id -= self._marker_offset
                filtered_markers.append(marker)
            return filtered_markers

        for marker in markers:
            if marker._id in self._cam.tag_sizes.keys():
                marker._id -= self._marker_offset
                filtered_markers.append(marker)

        return filtered_markers

    def annotated_frame_hook(self, frame: Frame, markers: List[Marker]) -> None:
        """
        During _detect annotate a copy of the frame and send it over MQTT.

        The image data is a base64 JPEG.
        """
        if self._mqtt_publish is None:
            return

        copied_frame = Frame(
            np.array(frame.grey_frame, copy=True),
            np.array(frame.colour_frame, copy=True),
        )
        annotated_frame = self._cam._annotate(copied_frame, markers)

        thread = Thread(
            name="Image send",
            target=self.background_encode_and_send,
            args=(annotated_frame.colour_frame,),
            daemon=True,
        )
        thread.start()

    def background_encode_and_send(self,  frame: NDArray[np.uint8]) -> None:
        """
        Handle converting a frame to a base64 bytestring and sending over MQTT.

        To be run as a thread target.
        """
        if self._mqtt_publish is None:
            return

        encoded_frame = self.base64_encode_frame(frame)
        if encoded_frame is None:
            return

        encoded_frame = b'data:image/jpeg;base64, ' + encoded_frame

        # The message is sent in a background thread
        try:
            self._mqtt_publish(
                'camera/annotated',
                encoded_frame,
            )
        except ValueError:
            pass

    def base64_encode_frame(self, frame: NDArray[np.uint8]) -> Optional[bytes]:
        """Convert image frame to base64 bytestring."""
        ret, image = cv2.imencode('.jpg', frame)
        if not ret:
            return None
        image_bytes = image.tobytes()
        return base64.b64encode(image_bytes)



"""Generic utility functions."""
from collections import deque
from math import hypot
from pathlib import Path
from typing import Deque, NamedTuple, Tuple, Union

import cv2
import numpy as np
from numpy.typing import NDArray

from .marker import Marker, PixelCoordinates

Resolution = Tuple[int, int]
CameraIntrinsic = Tuple[float, float, float, float]


class Frame(NamedTuple):
    """A tuple of original image and a greyscale version for fiducial detection."""

    grey_frame: NDArray
    colour_frame: NDArray

    @classmethod
    def from_colour_frame(cls, colour_frame: NDArray) -> 'Frame':
        """Load frame from a colour image in a numpy array."""
        grey_frame = cv2.cvtColor(colour_frame, cv2.COLOR_BGR2GRAY)

        return cls(
            grey_frame=grey_frame,
            colour_frame=colour_frame,
        )

    @classmethod
    def from_file(cls, filepath: Union[str, Path]) -> 'Frame':
        """Load an image file into the frame."""
        colour_frame = cv2.imread(str(filepath))

        return cls.from_colour_frame(colour_frame)


def annotate_text(
    frame: Frame,
    text: str,
    location: Tuple[int, int],
    text_scale: float = 0.5,
    text_colour: Tuple[int, int, int] = (255, 0, 0),
) -> Frame:
    """
    Add arbitrary text to a frame.

    text colour is in BGR format
    """
    for frame_type in frame:
        cv2.putText(
            frame_type,
            text,
            np.array(location, dtype=np.int32),
            cv2.FONT_HERSHEY_DUPLEX,
            text_scale,
            color=text_colour,  # in BGR
            thickness=2,
        )
    return frame


def load_calibration(calibration_file: Union[str, Path]) -> Tuple[Resolution, CameraIntrinsic]:
    """Load calibration data from opencv XML calibration file."""
    calibration_file = Path(calibration_file)

    if not calibration_file.exists():
        raise FileNotFoundError(f"Calibrations not found: {calibration_file}")

    storage = cv2.FileStorage(str(calibration_file), cv2.FILE_STORAGE_READ)

    resolution_node = storage.getNode("cameraResolution")
    width = int(resolution_node.at(0).real())
    height = int(resolution_node.at(1).real())

    camera_matrix = storage.getNode("cameraMatrix").mat()
    fx, fy = camera_matrix[0, 0], camera_matrix[1, 1]
    cx, cy = camera_matrix[0, 2], camera_matrix[1, 2]

    return (
        (width, height),
        (fx, fy, cx, cy),
    )


def normalise_marker_text(marker: Marker) -> float:
    """
    Calculate text scale factor so that text on marker is a reasonable size.

    Based on the distance between corner diagonals.
    """
    corners = marker.pixel_corners

    def line_distance(p1: PixelCoordinates, p2: PixelCoordinates) -> float:
        return hypot(p1.x - p2.x, p1.y - p2.y)

    return max(
        line_distance(corners[0], corners[2]),
        line_distance(corners[1], corners[3]),
    ) / 300


class RollingAverage:
    """A rolling average filter using a deque."""

    def __init__(self, points: int) -> None:
        self.data: Deque[float] = deque(maxlen=points)
        self.points = points

    def new_data(self, value: float) -> None:
        """
        Add a new sample to the filter.

        Once the buffer is full, the oldest sample is dropped.
        """
        self.data.append(value)

    def average(self) -> float:
        """Return the average of the stored samples."""
        active_points = [data for data in self.data if data is not None]

        return sum(active_points) / len(active_points)



# Copyright (c) 2021 Chris Reed
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Any, Optional

from usb.backend import IBackend
from pathlib import Path


@functools.lru_cache()
def get_library_path() -> Optional["Path"]: ...


def find_library(candidate: str) -> Optional[str]: ...


@functools.lru_cache()
def get_libusb1_backend() -> Optional["IBackend"]: ...


def find(*args: Any, **kwargs: Any) -> Any: ...



"""
Generate the debug images of the vision processing steps.

Provide an input image and generate the debug output
of the vision processing steps.
"""
import argparse
import logging
import os
from contextlib import contextmanager
from pathlib import Path

import cv2
from PIL import Image
from pyapriltags import Detector

LOGGER = logging.getLogger(__name__)

pnm_files = [
    "debug_preprocess.pnm",
    "debug_threshold.pnm",
    "debug_segmentation.pnm",
    "debug_clusters.pnm",
    "debug_quads_raw.pnm",
    "debug_quads_fixed.pnm",
    "debug_samples.pnm",
    "debug_output.pnm",
]

ps_files = [
    "debug_lines.ps",
    "debug_output.ps",
    "debug_quads.ps",
]


@contextmanager
def pushd(new_dir):
    """Enter a directory for the context and return to the previous on exiting."""
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)


def process_debug(preserve=True, collage=True, clean=False):
    """Convert debug outputs to PNG and optionally collage the images."""
    new_files = []

    for index, file in enumerate(pnm_files):
        LOGGER.info(f"Generating PNG for {file}.")
        with Image.open(file) as im:
            new_filename = f"{index}_{file.split('.')[0]}.png"
            im.save(new_filename)
            new_files.append(new_filename)

    if not preserve:
        for file in (ps_files + pnm_files):
            try:
                os.remove(file)
            except FileNotFoundError:
                pass

    if collage:
        create_collage(new_files, 'all.png')

        if clean:
            for file in new_files:
                try:
                    os.remove(file)
                except FileNotFoundError:
                    pass


def create_collage(files, out):
    """Create a collage of the debug images."""
    LOGGER.info("Generating debug collage.")
    img = Image.open(files[-1])
    width, height = img.size

    target_img = Image.new("RGB", (width*4, height*2))
    for k, png in enumerate(files):
        col, row = divmod(k, 4)
        img = Image.open(png)
        img.thumbnail((width, height))
        target_img.paste(img, (width*row, height*col))

    target_img.save(out)
    LOGGER.info("Generated debug collage.")


def main(args: argparse.Namespace):
    """Generate the debug images of the vision processing steps."""
    if not args.input_file.exists():
        LOGGER.fatal("Input file does not exist.")
        return

    detector = Detector(quad_decimate=1.0, debug=True)

    frame = cv2.imread(str(args.input_file), cv2.IMREAD_GRAYSCALE)

    if not args.output_dir.exists():
        args.output_dir.mkdir(parents=True)
    # change directory around the debug process
    with pushd(args.output_dir):
        results = detector.detect(frame)
        LOGGER.info(f"Found {len(results)} {'marker' if len(results)==1 else 'markers'}")

        process_debug(
            preserve=not args.cleanup,
            collage=args.collage,
            clean=args.collage_only)


def create_subparser(subparsers: argparse._SubParsersAction):
    """Vision_debug command parser."""
    parser = subparsers.add_parser(
        "vision_debug",
        description="Generate the debug images of the vision processing steps.",
        help="Generate the debug images of the vision processing steps.",
    )

    parser.add_argument("input_file", type=Path, help="The image to process.")
    parser.add_argument(
        "output_dir", type=Path, help="The directory to save the output files to.")
    parser.add_argument(
        '--collage', action='store_true',
        help="Generate the collage image 'all.png' of all the processing steps.")
    parser.add_argument(
        '--no-cleanup', action='store_false', dest='cleanup',
        help="Don't remove the interim PNM and PS files.")
    parser.add_argument(
        '--collage-only', action='store_true',
        help="Remove separate debug images, leaving only 'all.png'")

    parser.set_defaults(func=main)



"""
Classes for marker detections and various axis representations.

Setting the environment variable ZOLOTO_LEGACY_AXIS uses the axis that were
used in zoloto<0.9.0. Otherwise the conventional right-handed axis is used
where x is forward, y is left and z is upward.
"""
import os
from enum import Enum
from math import acos, atan2, pi
from typing import Any, Dict, List, NamedTuple, Tuple

import numpy as np
from numpy.linalg import norm as hypotenuse
from numpy.typing import NDArray
from pyapriltags import Detection
from pyquaternion import Quaternion


class MarkerType(Enum):
    """
    The available tag families.

    To support Apriltag 2 libraries use tag36h11.
    """

    APRILTAG_16H5 = 'tag16h5'
    APRILTAG_25H9 = 'tag25h9'
    APRILTAG_36H11 = 'tag36h11'
    APRILTAG_21H7 = 'tagCircle21h7'
    APRILTAG_49H12 = 'tagCircle49h12'
    APRILTAG_48H12 = 'tagCustom48h12'
    APRILTAG_41H12 = 'tagStandard41h12'
    APRILTAG_52H13 = 'tagStandard52h13'


class PixelCoordinates(NamedTuple):
    """
    Coordinates within an image made up from pixels.

    Floating point type is used to allow for subpixel detected locations
    to be represented.

    :param float x: X coordinate
    :param float y: Y coordinate
    """

    x: float
    y: float


class CartesianCoordinates(NamedTuple):
    """
    A 3 dimesional cartesian coordinate in the standard right-handed cartesian system.

    Origin is at the camera.

    The X axis extends directly away from the camera. Zero is at the camera.
    Increasing values indicate greater distance from the camera.

    The Y axis is horizontal relative to the camera's perspective, i.e: right
    to left within the frame of the image. Zero is at the centre of the image.
    Increasing values indicate greater distance to the left.

    The Z axis is vertical relative to the camera's perspective, i.e: down to
    up within the frame of the image. Zero is at the centre of the image.
    Increasing values indicate greater distance above the centre of the image.

    More information: https://w.wiki/5zbE

    Legacy:
    The X axis is horizontal relative to the camera's perspective, i.e: left &
    right within the frame of the image. Zero is at the centre of the image.
    Increasing values indicate greater distance to the right.

    The Y axis is vertical relative to the camera's perspective, i.e: up & down
    within the frame of the image. Zero is at the centre of the image.
    Increasing values indicate greater distance below the centre of the image.

    The Z axis extends directly away from the camera. Zero is at the camera.
    Increasing values indicate greater distance from the camera.

    These match traditional cartesian coordinates when the camera is facing
    upwards.


    :param float x: X coordinate, in millimeters
    :param float y: Y coordinate, in millimeters
    :param float z: Z coordinate, in millimeters
    """

    x: float
    y: float
    z: float

    @classmethod
    def from_tvec(cls, x: float, y: float, z: float):
        """
        Convert coordinate system to standard right-handed cartesian system.

        The pose estimation coordinate system has the origin at the camera center.
        Also converts units to millimeters.

        :param float x: The x-axis is to the right in the image taken by the camera.
        :param float y: The y-axis is down in the image taken by the camera.
        :param float z: The z-axis points from the camera center out the camera lens.
        """
        if os.environ.get('ZOLOTO_LEGACY_AXIS'):
            return cls(x=x * 1000, y=y * 1000, z=z * 1000)
        else:
            return cls(x=z * 1000, y=-x * 1000, z=-y * 1000)


class SphericalCoordinate(NamedTuple):
    """
    A 3 dimesional spherical coordinate location.

    The convential spherical coordinate in mathematical notation where θ is
    a rotation around the vertical axis and φ is measured as the angle from
    the vertical axis.
    More information: https://mathworld.wolfram.com/SphericalCoordinates.html

    :param float distance: Radial distance from the origin, in millimeters.
    :param float theta: Azimuth angle, θ, in radians. This is the angle from
        directly in front of the camera to the vector which points to the
        location in the horizontal plane. A positive value indicates a
        counter-clockwise rotation. Zero is at the centre of the image.
    :param float phi: Polar angle, φ, in radians. This is the angle "down"
        from the vertical axis to the vector which points to the location.
        Zero is directly upward.
    """

    distance: int
    theta: float
    phi: float

    @property
    def rot_x(self) -> float:
        """
        Rotation around the x-axis.

        Conventional:  This is unused.
        Legacy: A rotation up to down around the camera, in radians. Values
                increase as the marker moves towards the bottom of the image.
                A zero value is halfway up the image.
        """
        if os.environ.get('ZOLOTO_LEGACY_AXIS'):
            return self.phi - (pi / 2)
        else:
            raise AttributeError(
                "That axis is not available in the selected coordinate system.")

    @property
    def rot_y(self) -> float:
        """
        Rotation around the y-axis.

        Conventional: A rotation up to down around the camera, in radians.
                      Values increase as the marker moves towards the bottom
                      of the image. A zero value is halfway up the image.
        Legacy: A rotation left to right around the camera, in radians. Values
                increase as the marker moves towards the right of the image.
                A zero value is on the centerline of the image.
        """
        if os.environ.get('ZOLOTO_LEGACY_AXIS'):
            return -self.theta
        else:
            return self.phi - (pi / 2)

    @property
    def rot_z(self) -> float:
        """
        Rotation around the z-axis.

        Conventional: A rotation right to left around the camera, in radians.
                      Values increase as the marker moves towards the left of
                      the image. A zero value is on the centerline of the
                      image.
        Legacy: This is unused.
        """
        if os.environ.get('ZOLOTO_LEGACY_AXIS'):
            raise AttributeError(
                "That axis is not available in the selected coordinate system.")
        else:
            return self.theta

    @classmethod
    def from_tvec(cls, x: float, y: float, z: float):
        """
        Convert coordinate system to standard right-handed cartesian system.

        The pose estimation coordinate system has the origin at the camera center.

        :param float x: The x-axis is to the right in the image taken by the camera.
        :param float y: The y-axis is down in the image taken by the camera.
        :param float z: The z-axis points from the camera center out the camera lens.
        """
        _x, _y, _z = z, -x, -y

        dist = hypotenuse([_x, _y, _z])
        theta = atan2(_y, _x)
        phi = acos(_z / dist)
        return cls(int(dist * 1000), theta, phi)


ThreeTuple = Tuple[float, float, float]
RotationMatrix = Tuple[ThreeTuple, ThreeTuple, ThreeTuple]


class Orientation:
    """The orientation of an object in 3-D space."""

    __MARKER_ORIENTATION_CORRECTION = Quaternion(matrix=np.array([
        [1, 0, 0],
        [0, -1, 0],
        [0, 0, -1],
    ]))
    __ZOLOTO_LEGACY_ORIENTATION = Quaternion(matrix=np.array([
        [-1, 0, 0],
        [0, 1, 0],
        [0, 0, -1],
    ]))

    def __init__(self, rotation_matrix: NDArray, aruco_orientation: bool = True):
        """
        Construct a quaternion given the rotation matrix in the camera's coordinate system.

        More information:
        https://en.wikipedia.org/wiki/Rotation_matrix#In_three_dimensions
        """
        # Calculate the quaternion of the rotation in the camera's coordinate system
        initial_rot = Quaternion(matrix=rotation_matrix)
        # remap quaternion to global coordinate system from the token's perspective
        quaternion = Quaternion(
            initial_rot.w, -initial_rot.z, -initial_rot.x, initial_rot.y,
        )
        if aruco_orientation:
            # Rotate the quaternion so 0 roll is a marker the correct way up
            quaternion *= self.__MARKER_ORIENTATION_CORRECTION

        self.__rotation_matrix = quaternion.rotation_matrix

        if os.environ.get('ZOLOTO_LEGACY_AXIS'):
            self._quaternion = initial_rot * self.__ZOLOTO_LEGACY_ORIENTATION
        else:
            self._quaternion = quaternion

        self._yaw_pitch_roll = quaternion.yaw_pitch_roll

    @property
    def rot_x(self) -> float:
        """
        Get rotation angle around X axis in radians.

        The roll rotation with zero as the April Tags marker reference point
        at the top left of the marker.

        Legacy: The inverted pitch rotation with zero as the marker facing
                directly away from the camera and a positive rotation being
                downward.
                The practical effect of this is that an April Tags marker
                facing the camera square-on will have a value of ``pi`` (or
                equivalently ``-pi``).
        """
        return self.yaw_pitch_roll[2]

    @property
    def rot_y(self) -> float:
        """
        Get rotation angle around Y axis in radians.

        The pitch rotation with zero as the marker facing the camera square-on
        and a positive rotation being upward.

        Legacy: The inverted yaw rotation with zero as the marker facing the
                camera square-on and a positive rotation being
                counter-clockwise.
        """
        return self.yaw_pitch_roll[1]

    @property
    def rot_z(self) -> float:
        """
        Get rotation angle around Z axis in radians.

        The yaw rotation with zero as the marker facing the camera square-on
        and a positive rotation being clockwise.

        Legacy: The roll rotation with zero as the marker facing the camera
                square-on and a positive rotation being clockwise.
        """
        return self.yaw_pitch_roll[0]

    @property
    def yaw(self) -> float:
        """
        Get yaw of the marker, a rotation about the vertical axis, in radians.

        Positive values indicate a rotation clockwise from the perspective of
        the marker.

        Zero values have the marker facing the camera square-on.
        """
        return self._yaw_pitch_roll[0]

    @property
    def pitch(self) -> float:
        """
        Get pitch of the marker, a rotation about the transverse axis, in radians.

        Positive values indicate a rotation upwards from the perspective of the
        marker.

        Zero values have the marker facing the camera square-on.
        """
        return self._yaw_pitch_roll[1]

    @property
    def roll(self) -> float:
        """
        Get roll of the marker, a rotation about the longitudinal axis, in radians.

        Positive values indicate a rotation clockwise from the perspective of
        the marker.

        Zero values have the marker facing the camera square-on.
        """
        return self._yaw_pitch_roll[2]

    @property
    def yaw_pitch_roll(self) -> ThreeTuple:
        """
        Get the equivalent yaw-pitch-roll angles.

        Specifically intrinsic Tait-Bryan angles following the z-y'-x'' convention.
        """
        return self._quaternion.yaw_pitch_roll

    @property
    def rotation_matrix(self) -> RotationMatrix:
        """
        Get the rotation matrix represented by this orientation.

        Returns:
            A 3x3 rotation matrix as a tuple of tuples.
        """
        return self.__rotation_matrix.tolist()

    @property
    def quaternion(self) -> Quaternion:
        """Get the quaternion represented by this orientation."""
        return self._quaternion

    def __repr__(self) -> str:
        return "Orientation(rot_x={}, rot_y={}, rot_z={})".format(
            self.rot_x, self.rot_y, self.rot_z,
        )


class Marker:
    """Wrapper of a marker detection with axis and rotation calculated."""

    def __init__(
        self,
        marker: Detection,
        *,
        aruco_orientation: bool = True,
    ):
        self.marker = marker

        self.__marker_type = MarkerType(marker.tag_family.decode('utf-8'))
        self._id = marker.tag_id
        self.__pixel_center = PixelCoordinates(*marker.center.tolist())
        self._pixel_corners = marker.corners.tolist()
        self.__size = int(marker.tag_size * 1000) if marker.tag_size is not None else 0
        self._tvec = marker.pose_t
        self._rvec = marker.pose_R
        self.__aruco_orientation = aruco_orientation

        if self._tvec is not None:
            self.__distance = int(hypotenuse(self._tvec) * 1000)
        else:
            self.__distance = 0

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} id={self.id} size={self.size} "
            f"type={self.marker_type.name} distance={self.distance}>")

    @property  # noqa: A003
    def id(self) -> int:  # noqa: A003
        """The marker id number."""
        return self._id

    @property
    def size(self) -> int:
        """The size of the detected marker in millimeters."""
        return self.__size

    @property
    def marker_type(self) -> MarkerType:
        """The family of the detected marker, likely tag36h11."""
        return self.__marker_type

    @property
    def pixel_corners(self) -> List[PixelCoordinates]:
        """The pixels of the corners of the marker in the image."""
        return [
            PixelCoordinates(x, y)
            for x, y in self._pixel_corners
        ]

    @property
    def pixel_centre(self) -> PixelCoordinates:
        """The pixel location of the center of the marker in the image."""
        return self.__pixel_center

    @property
    def distance(self) -> int:
        """The distance between the marker and camera, in millimeters."""
        if self._tvec is not None and self._rvec is not None:
            return self.__distance
        return 0

    @property
    def orientation(self) -> Orientation:
        """The marker's orientation."""
        if self._rvec is not None:
            return Orientation(self._rvec, aruco_orientation=self.__aruco_orientation)
        raise RuntimeError("This marker was detected with an uncalibrated camera")

    @property
    def spherical(self) -> SphericalCoordinate:
        """The spherical coordinates of the marker's location relative to the camera."""
        if self._tvec is not None:
            return SphericalCoordinate.from_tvec(*self._tvec.flatten().tolist())
        raise RuntimeError("This marker was detected with an uncalibrated camera")

    @property
    def cartesian(self) -> CartesianCoordinates:
        """The cartesian coordinates of the marker's location relative to the camera."""
        if self._tvec is not None:
            return CartesianCoordinates.from_tvec(*self._tvec.flatten().tolist())
        raise RuntimeError("This marker was detected with an uncalibrated camera")

    def as_dict(self) -> Dict[str, Any]:
        """The marker data as a dict."""
        marker_dict = {
            "id": self._id,
            "size": self.__size,
            "pixel_corners": self._pixel_corners,
        }
        if self._tvec is not None and self._rvec is not None:
            marker_dict.update(
                {"rvec": self._rvec.tolist(), "tvec": self._tvec.tolist()},
            )
        return marker_dict
